/* 

   Composite linkage disequilibrium is a statistical measure of
   association across loci, when haplotype phase is not known. See
   Weir and Cockerham, 1989; Weir 1996 pp. 94, 125.

   Delta = n / N - 2 p_{A_1} p_{B_1}

   where
   
   N: number of individuals
   p_{A_1}: frequency of allele A_1 at snp A
   p_{B_1}: frequency of an allele B_1 at snp B
   n: 2 X_{A_1A_1, B_1B_1} + X_{A_1A_1, B_1B_2} +
       X_{A_1A_2, B_1B_1} + (1/2) X_{A_1A_2, B_1B_2}

   WEIR, B., 1996 Genetic Data Analysis II. Sinauer Associates,
   Sunderland, MA.
       
   WEIR, B. S., and C. C. COCKERHAM, 1989 Complete characterization of
   disequilibrium at two loci, pp. 86–110 in Mathematical Evolutionary
   Theory, edited by M. W. FELDMAN. Princeton University Press,
   Princeton, NJ.

 */

#include "composite_linkage_disequilibrium.h"

void
composite_linkage_disequilibrium(unsigned char *snp, /* matrix indiv x snp */
				 int *n_ind, /* # individuals */
				 int *n_snp, /* # snps */
				 int *width, /* adjacent snp window */
				 double *delta) /* result */
{
    /* convenience and index variables */
    int m = n_snp[0], n = n_ind[0], w = width[0];
    int i, j1, j2, k;

    /* intermediate values */
    int X[9];			/* 9 genotype frequencies */
    double p_A1, p_B1, n_A1_B1, N;

    /* for each combination of snps... */
    for (j1 = 0; j1 < (m - w) * n; j1 += n)
	for (j2 = j1 + n; j2 <= j1 + n * w; j2 += n) 
	{
	    /* ...tally 9 genotype frequencies */
	    for (k = 0; k < 9; ++k)
		X[k] = 0;
	    for (i = 0; i < n; ++i)
		if (snp[j1 + i] & snp[j2 + i])
		    X[ (snp[j1 + i] - 1) * 3 + snp[j2 + i] - 1] += 1;
	    
	    /* ...and calculate composite linkage disequilibrium */
	    p_A1 = X[0] + X[1] + X[2] + (X[3] + X[4] + X[5]) / 2;
	    p_B1 = X[0] + X[3] + X[5] + (X[1] + X[4] + X[6]) / 2;
	    n_A1_B1 = 2 * X[0] + X[1] + X[3] + X[4] / 2;
	    for (k = 0, N = 0; k < 9; ++k)
		N += X[k];
	    *delta++ = n_A1_B1 / N - 2 * p_A1 * p_B1;
	}
}

SEXP
composite_linkage_disequilibrium_1(SEXP snp_r, SEXP width_r)
{
    /* check incoming types */
    if (!IS_INTEGER(width_r) || 1 != Rf_length(width_r))
	Rf_error("'width' must be integer(1)");
    if (!IS_RAW(snp_r) || 
	2 != Rf_length(getAttrib(snp_r, R_DimSymbol)))
    {
	Rf_error("'snp' must be matrix of raw()");
    }

    /* retrieve inputs */
    int width = INTEGER(width_r)[0],
	n_ind = INTEGER(getAttrib(snp_r, R_DimSymbol))[0],
	n_snp = INTEGER(getAttrib(snp_r, R_DimSymbol))[1];

    /* allocate memory for return, set dims */
    SEXP delta_r, dim_r;
    PROTECT(delta_r = allocVector(REALSXP, n_snp * (n_snp - width)));
    PROTECT(dim_r = allocVector(INTSXP, 2));
    REAL(dim_r)[0] = n_snp;
    REAL(dim_r)[1] = n_snp * (n_snp - width);
    setAttrib(delta_r, R_DimSymbol, dim_r);

    /* do the calculation */
    composite_linkage_disequilibrium(RAW(snp_r), 
				     &n_ind, &n_snp, &width,
				     REAL(delta_r));

    /* unprotect and return */    
    UNPROTECT(2);
    return delta_r;
}
